# $Id: yahoo.rb,v 1.16 2005/03/24 22:03:06 ianmacd Exp $
#

require 'finance/quote/yahoo/asia'
require 'finance/quote/yahoo/australia'
require 'finance/quote/yahoo/europe'
require 'finance/quote/yahoo/usa'
require 'net/http'
require 'csv'

module Finance

  # This class implements financial quote operations.
  #
  class Quote

    # This module interfaces with Yahoo's financial data services. It is not
    # intended for direct use by the user, but instead implements methods for
    # use by other Yahoo-related modules.
    #
    # On the other hand, this module may be of interest to you if you are
    # intending to implement an extension module that retrieves information
    # via Yahoo.
    # 
    module Yahoo

      Finance::Quote::register_plugin( self,
				       { :yahoo_request => :yahoo_request } )

      # This is the maximum number of stocks we'll batch into a single
      # operation. If this gets too big (> 50 or thereabouts), things will
      # break, because some proxies and/or webservers cannot handle very
      # large URLs.
      #
      MAX_REQUEST_SIZE = 40

      # Yahoo encodes the desired fields as 1 or 2 character strings in the
      # URL. These are recorded below, along with their corresponding field
      # names.
      #
      LABELS = %w[ symbol name last date time net p_change volume bid ask
		   close open day_range year_range eps pe div_date div
		   div_yield cap ex_div avg_vol ].map { |label| label.to_sym }

      FIELD_ENCODING = %w[ s n l1 d1 t1 c1 p2 v b a
			   p o m w e r r1 d y j1 q a2 ]

      # Yahoo uses a suffix on the stock symbol to denote the exchange on
      # which the stock is traded. Use this suffix to map from the stock
      # symbol to the currency in which its prices are reported.
      #
      CURRENCY_TAGS = {
	:US  => 'USD',	# USA		AMEX, Nasdaq, NYSE
  	:OB  => 'USD',	# USA		OTC Bulletin Board
  	:PK  => 'USD',	# USA		Pink Sheets
	:BA  => 'ARS',	# Argentina	Buenos Aires
	:VA  => 'EUR',	# Austria	Vienna
  	:AX  => 'AUD',	# Australia
	:SA  => 'BRL',	# Brazil	Sao Paolo
  	:TO  => 'CAD',	# Canada	Toronto
  	:V   => 'CAD',	# 		Toronto Venture
	:SN  => 'CLP',	# Chile		Santiago
	:SS  => 'CNY',	# China		Shanghai
	:SZ  => 'CNY',	# 		Shenzhen
  	:CO  => 'DKK',	# Denmark	Copenhagen
  	:PA  => 'EUR',	# France	Paris
  	:BE  => 'EUR',	# Germany	Berlin
  	:BM  => 'EUR',	# 		Bremen
  	:D   => 'EUR',	# 		Dusseldorf
  	:F   => 'EUR',	# 		Frankfurt
  	:H   => 'EUR',	# 		Hamburg
  	:HA  => 'EUR',	# 		Hanover
  	:MU  => 'EUR',	# 		Munich
  	:SG  => 'EUR',	# 		Stuttgart
  	:DE  => 'EUR',	# 		XETRA
	:HK  => 'HKD',	# Hong Kong
	:BO  => 'INR',	# India		Bombay
	:CL  => 'INR',	# 		Calcutta
	:NS  => 'INR',	# 		National Stock Exchange
	:JK  => 'IDR',	# Indonesia	Jakarta
	:TA  => 'ILS',	# Israel	Tel Aviv
  	:MI  => 'EUR',	# Italy		Milan
	:KS  => 'KRW',	# Korea		Stock Exchange
	:KQ  => 'KRW',	# 		KOSDAQ
	:KL  => 'MYR',	# Malaysia	Kuala Lampur
	:MX  => 'MXP',	# Mexico
	:NZ  => 'NZD',	# New Zealand
  	:AS  => 'EUR',	# Netherlands	Amsterdam
  	:OS  => 'NOK',	# Norway	Oslo
	:LM  => 'PEN',	# Peru		Lima
	:SI  => 'SGD',	# Singapore
	:BC  => 'EUR',	# Spain		Barcelona
	:BI  => 'EUR',	# 		Bilbao
	:MF  => 'EUR',	# 		Madrid Fixed Income
	:MC  => 'EUR',	# 		Madrid SE CATS
	:MA  => 'EUR',	# 		Madrid
  	:ST  => 'SEK',	# Sweden	Stockholm
	:TW  => 'TWD',	# Taiwan	Taiwan Stock Exchange
	:TWO => 'TWD',	# 		OTC
	:BK  => 'THB',	# Thailand	Thailand Stock Exchange
	:TH  => 'THB',	# 		??? from asia.rb, (in Thai Baht)
  	:L   => 'GBP' 	# United Kingdom London
      }

      # This method takes the base URL to use and an Array of symbols to look
      # up. If a third argument is given, it will act as a suffix to be
      # appended to the stocks in order to obtain the correct information.
      # This method relies on the fact that the various Yahoo sites all work
      # in the same way.
      #
      def yahoo_request( base_url, orig_symbols, suffix='' )

	base_url = base_url.dup
	orig_symbols = orig_symbols.dup
	info = QuoteData.new

	# Generate a suitable URL, now all it needs is the ticker symbols.
	base_url << '?f=' << FIELD_ENCODING.join( '' ) << '&e=.csv&s='

	while ( symbols = orig_symbols.slice!( 0, MAX_REQUEST_SIZE ) ) != []

	  symbols.map! { |symbol| url_encode( symbol ) }

	  # By pushing an extra symbol onto our array, we can be sure that
	  # everything ends up with the correct suffix in the join() below.
	  #
	  symbols << ''

	  url = base_url + symbols.join( suffix + '+' )
	  url.chop!

	  yahoo = URI.parse( url )
	  host = yahoo.host
	  path = yahoo.path
	  query = yahoo.query

	  Net::HTTP::Proxy( *Finance::proxy_data ).start( host, 80 ) do |http|

	    if $DEBUG
	      $stderr.printf( "Getting http://%s%s?%s ...\n",
			      host, path, query )
	    end

	    response = http.get( path + '?' + query )
	    body = response.body

	    # We now have the data. Put it in the Hash.
	    body.split( /\r?\n/ ).each do |line|
	      fields = CSV.parse_line( line )
	      symbol = fields[0]
	      exchange = $1 if symbol =~ /\.([A-Z]+)/

	      # Strip out suffixes.
	      symbol[-(suffix.length),suffix.length] = ''

	      # Load all the returned fields into our Hash.
	      info[symbol] ||= {}
	      LABELS.each_index do |idx|
		info[symbol][LABELS[idx]] = fields[idx]
	      end

	      # Yahoo returns a line filled with N/A values if we look up a
	      # non-existent symbol. Perl's Finance::Quote claims that the
	      # date field will never be defined for a bogus stock, but I have
	      # found this to be false. For example, looking up *BOGUS* via
	      # _yahoo_europe_ returns a filled date.
	      #
	      # Instead, we'll try to use _day_range_ in combination with
	      # _avg_vol_ for now and hope that this pairing is always a good
	      # indicator.
	      #
	      if info[symbol][:day_range] == 'N/A - N/A' &&
		 info[symbol][:avg_vol].to_i == 0
		info[symbol][:success] = false
		info[symbol][:error_msg] = 'Stock look-up failed.'
		next
	      end

	      # Look-up succeeded, so continue.
	      info[symbol][:success] = true
	      info[symbol][:price] = info[symbol][:last]

	      # Remove spurious percentage signs in p_change.
	      info[symbol][:p_change].delete!( '%' )

	      # Extract the high and low values from the day-range, if
	      # available.
	      if info[symbol][:day_range] =~ /^"?\s*(\S+)\s*-\s*(\S+)"?$/
		info[symbol][:low] = $1
		info[symbol][:high] = $2
	      end

	      # Determine the currency from the exchange name. Symbols
	      # without an exchange are in USD. Symbols starting with a '^'
	      # are always indexes, so they don't have a currency.
	      if exchange
		info[symbol][:currency] = CURRENCY_TAGS[exchange.to_sym]
	      elsif symbol[0,1] != '^'
		info[symbol][:currency] = 'USD'
	      end

	      # Convert prices, if needed. For example, London sources return
	      # in pence. We'd like them to return in pounds (divide by 100).
	      if exchange == 'L'
		Finance::Currency::DEFAULT_CURRENCY_FIELDS.each do |field|
		  next unless info[symbol][field]

		  info[symbol][field] =
		    Finance::Quote.scale_field( info[symbol][field], 0.01 )
		end

		# Other exchanges here as needed.
	      end
	    end
	  end

	  # Remove keys whose value is N/A. Strip out HTML, too.
	  info.keys.each do |symbol|
	    info[symbol].keys.each do |field|
	      next unless info[symbol][field].kind_of?( String )
	      
	      info[symbol][field].gsub!( /<[^>]*>/, '' )
	      info[symbol][field].sub!( /&nbsp;.*$/, '' )
	      info[symbol].delete( field ) if info[symbol][field] == 'N/A'
	    end
	  end

	  convert( info, Finance::Currency::DEFAULT_CURRENCY_FIELDS )

	  yield info if block_given?

	  return info

	end
      end

      private :yahoo_request

      # Encode a string, such that it is suitable for HTTP transmission.
      #
      def url_encode( string )
        # shamelessly plagiarised from Wakou Aoyama's cgi.rb
        string.gsub( /([^ a-zA-Z0-9_.-]+)/n ) do
          '%' + $1.unpack( 'H2' * $1.size ).join( '%' ).upcase
        end.tr( ' ', '+' )
      end

      private :url_encode

    end
  end
end
